"""Classes, enums, and misc data containers."""

import argparse
import logging
from enum import Enum
from pathlib import Path
from typing import List, Optional, Tuple

import attr

LOG = logging.getLogger(__name__)


class Position(Enum):
    """Predefined locations of text watermark."""

    TOP_LEFT = "top-left"
    TOP_RIGHT = "top-right"
    BOTTOM_RIGHT = "bottom-right"
    BOTTOM_LEFT = "bottom-left"

    def __str__(self):
        """Return enum value in lowercase."""
        return self.value.lower()

    def __repr__(self):
        """Return string representation of enum."""
        return str(self)

    @staticmethod
    def argparse(s):
        """Parse string values from CLI into Position.

        Parameters
        ----------
        - `s` String value to match against enum attribute

        """
        vals = {x.value.lower(): x for x in list(Position)}
        try:
            return vals[s.lower()]
        except KeyError:
            return s


@attr.s(auto_attribs=True)
class Config:
    """Store options from config file and command line."""

    input_file: Optional[Path] = None
    output_file: Optional[Path] = None
    verbosity: int = 0
    suffix: Optional[str] = None
    force: bool = False
    no_op: bool = False
    pct_scale: float = 0
    width: int = 0
    height: int = 0
    keep_exif: bool = False
    watermark_image: Optional[Path] = None
    watermark_rotation: int = 0
    watermark_opacity: float = 0
    watermark_position: Optional[Position] = None
    watermark_scale: float = 0
    text: Optional[str] = None
    text_copyright: Optional[str] = None
    text_rotation: int = 0
    text_opacity: float = 0
    text_position: Optional[Position] = None
    text_scale: int = 0
    jpg_quality: int = 0

    @staticmethod
    def from_args(args: argparse.Namespace):
        """Create Config instance from file and command line args."""
        cfg = Config()
        cfg.input_file = args.input
        cfg.output_file = args.output
        cfg.verbosity = args.verbosity
        cfg.suffix = args.suffix
        cfg.force = args.force
        cfg.no_op = args.no_op
        cfg.pct_scale = args.pct_scale
        cfg.width = args.width
        cfg.height = args.height
        cfg.keep_exif = args.keep_exif
        cfg.watermark_image = args.watermark_image
        cfg.watermark_rotation = args.watermark_rotation
        cfg.watermark_opacity = args.watermark_opacity
        cfg.watermark_position = args.watermark_position
        cfg.watermark_scale = args.watermark_scale
        cfg.text = args.text
        cfg.text_copyright = args.text_copyright
        cfg.text_rotation = args.text_rotation
        cfg.text_opacity = args.text_opacity
        cfg.text_position = args.text_position
        cfg.text_scale = args.text_scale
        cfg.jpg_quality = args.jpg_quality
        return cfg

    def as_dict(self, exclude_attrs: List[str] = []) -> dict:
        """Return dict representation of object.

        Parameters
        ----------
        - `exclude_attrs` List of attribute names to exclude from dict output

        Returns
        -------
        - dict

        """
        return attr.asdict(
            self, filter=lambda attr, value: attr.name not in exclude_attrs
        )


@attr.s(auto_attribs=True)
class ImageSize:
    """Pixel dimensions of image."""

    width: int = 0
    height: int = 0

    def __str__(self):
        """Return string representation, e.g.: width x height px."""
        return f"{self.width} x {self.height} px"

    def __repr__(self):
        """Return string representation, same as __str__."""
        return self.__str__

    @property
    def area(self) -> int:
        """Pixel area of image."""
        return self.width * self.height


@attr.s(auto_attribs=True)
class Context:
    """Keep track of details of processing for later reporting."""

    orig_size: ImageSize = ImageSize()
    new_size: ImageSize = ImageSize()
    watermark_size: ImageSize = ImageSize()
    orig_file_size: int = 0
    new_file_size: int = 0
    orig_dpi: Tuple[int, int] = (0, 0)
    new_dpi: Tuple[int, int] = (0, 0)
    image_buffer: Optional[bytes] = None
    orig_exif: Optional[dict] = None
    time_start: float = 0
    time_end: float = 0

    def as_dict(self, exclude_attrs: List[str] = ["image_buffer", "orig_exif"]) -> dict:
        """Return dict representation of object.

        Parameters
        ----------
        - `exclude_attrs` List of attribute names to exclude from dict output

        Returns
        -------
        - dict

        """
        return attr.asdict(
            self, filter=lambda attr, value: attr.name not in exclude_attrs
        )



"""Custom exceptions for pyimgtool."""


class ImageSizeError(Exception):
    """Raised when the supplied image does not fit the intial size requirements."""

    def __init__(self, actual_size, required_size):
        """Pass in variables for exception message."""
        self.message = "Image is too small, Image size : %s, Required size : %s" % (
            actual_size,
            required_size,
        )
        self.actual_size = actual_size
        self.required_size = required_size

    def __str__(self):
        """Return string representation."""
        return repr(self.message)



"""Watermark image with text or another image."""

import logging
import os
from pathlib import PurePath
from datetime import datetime

from PIL import Image, ImageDraw, ImageFont, ImageStat

from pyimgtool import resize
from pyimgtool.data_structures import Config, Context, ImageSize

LOG = logging.getLogger(__name__)


def with_image(im: Image, cfg: Config, ctx: Context) -> Image:
    """Watermark with image according to Config.

    Parameters
    ----------
    - `im` PIL Image
    - `cfg` Config object
    - `ctx` Context object

    """
    if cfg.watermark_image is None:
        LOG.error("Missing watermark_image in cfg")
        return im
    watermark_image = Image.open(os.path.expanduser(cfg.watermark_image)).convert(
        "RGBA"
    )
    LOG.info("Watermark: %s", watermark_image.size)
    watermark_ratio = watermark_image.height / im.height
    LOG.info("Watermark size ratio: %.4f", watermark_ratio)
    if watermark_ratio > cfg.watermark_scale:
        LOG.debug(
            "Resizing watermark from %.4f to %.4f scale",
            watermark_ratio,
            cfg.watermark_scale,
        )
        watermark_image = resize.resize_height(
            watermark_image,
            (int(im.width * cfg.watermark_scale), int(im.height * cfg.watermark_scale)),
        )
        LOG.debug("New watermark dims: %s", watermark_image.size)
    ctx.watermark_size = ImageSize(watermark_image.width, watermark_image.height)
    mask = watermark_image.split()[3].point(lambda i: i * cfg.watermark_opacity)
    pos = (
        (im.width - watermark_image.width - 25),
        (im.height - watermark_image.height - 25),
    )
    im.paste(watermark_image, pos, mask)
    return im


def with_text(im: Image, cfg: Config, ctx: Context) -> Image:
    """Watermark with text if program option is supplied.

    If text is equal to 'copyright', the exif data will be read
    (if available) to attempt to determine copyright date based on
    date photo was taken.

    Parameters
    ----------
    - `im` PIL Image
    - `cfg` Config object
    - `ctx` Context object

    """
    if cfg.text is None and cfg.text_copyright is None:
        LOG.error("Missing text or copyright text in cfg")
        return im
    if cfg.text_copyright is not None:
        photo_dt = datetime.now()
        if ctx.orig_exif is not None:
            try:
                photo_dt = datetime.strptime(
                    ctx.orig_exif["Exif"]["DateTimeOriginal"].decode("utf-8"),
                    "%Y:%m:%d %H:%M:%S",
                )
            except KeyError:
                pass
        copyright_year = photo_dt.strftime("%Y")
        cfg.text = f"© {copyright_year} {cfg.text_copyright}"
        LOG.info("Using copyright text: %s", cfg.text)
        LOG.info("Photo date from exif: %s", photo_dt)
    layer = Image.new("RGBA", (im.width, im.height), (255, 255, 255, 0))

    font_size = 1  # starting size
    offset_x = 10
    offset_y = 10

    try:
        cwd = PurePath(os.path.dirname(__file__))
        font_path = str(PurePath.joinpath(cwd, "fonts", "SourceSansPro-Regular.ttf"))
        font = ImageFont.truetype(font=font_path, size=font_size)
    except OSError:
        LOG.error("Could not find font '%s', aborting text watermark", font_path)
        return im

    LOG.debug("Found font '%s'", font_path)
    while font.getsize(cfg.text)[0] < cfg.text_scale * im.width:
        # iterate until text size is >= text_scale
        font_size += 1
        font = ImageFont.truetype(font=font_path, size=font_size)

    if font.getsize(cfg.text)[0] > cfg.text_scale * im.width:
        font_size -= 1
        font = ImageFont.truetype(font=font_path, size=font_size)

    text_width, text_height = font.getsize(cfg.text)
    stats = ImageStat.Stat(
        im.crop((offset_x, offset_y, text_width, im.height - text_height)).convert("L")
    )
    LOG.debug(
        "Final text dims: %d x %d px; Font size: %d", text_width, text_height, font_size
    )
    LOG.debug("Text region stats: mean=%f, rms=%f", stats.mean[0], stats.rms[0])
    d = ImageDraw.Draw(layer)
    opacity = int(round((cfg.text_opacity * 255)))
    LOG.info("Text opacity: %d/255", opacity)

    text_fill = 255, 255, 255, opacity
    if stats.mean[0] > 60:
        text_fill = 0, 0, 0, opacity

    d.text((offset_x, offset_y), cfg.text, font=font, fill=text_fill)
    out = Image.alpha_composite(im.convert("RGBA"), layer)
    return out.convert("RGB")



"""Resize and watermark images."""

import logging
import sys
from sty import fg, ef, rs
from pprint import pformat
from time import perf_counter
from pathlib import Path

from pyimgtool.args import parse_args
from pyimgtool.data_structures import Config
from pyimgtool.image import process_image
from pyimgtool.utils import get_summary_report

logging.basicConfig(level=logging.WARNING)
LOG = logging.getLogger(__name__)


def main():
    """Process image based on config file and command-line arguments."""
    time_start = perf_counter()
    cfg = Config.from_args(parse_args(sys.argv[1:]))
    log_level = 0
    try:
        log_level = (0, 20, 10)[cfg.verbosity]
    except IndexError:
        log_level = 10
    loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]
    # set level for all loggers
    for l in loggers:
        l.setLevel(log_level)

    LOG.debug("Runtime config:\n%s", pformat(cfg.as_dict(), indent=2))
    ctx = process_image(cfg)
    ctx.time_start = time_start
    LOG.debug(
        "Image Context:\n%s",
        pformat(ctx.as_dict(exclude_attrs=["image_buffer", "orig_exif"]), indent=2),
    )

    if not cfg.no_op:
        if not ctx.image_buffer:
            LOG.critical("Image buffer cannot be None")
            raise ValueError("Image buffer is None")
        LOG.info("Saving buffer to %s", cfg.output_file)

        # Create output dir if it doesn't exist
        out_path = Path(cfg.output_file)
        if out_path.exists():
            # output file exists
            if not cfg.force:
                print(
                    fg.red
                    + ef.bold
                    + f"Error: file '{out_path}' exists; use -f option to force overwrite."
                    + rs.all,
                    file=sys.stderr,
                )
                return
        out_path.parent.mkdir(parents=True, exist_ok=True)

        with out_path.open("wb") as f:
            f.write(ctx.image_buffer)
    else:
        print(fg.li_magenta + "***Displaying Results Only***" + fg.rs)

    ctx.time_end = perf_counter()
    print(*get_summary_report(cfg, ctx), sep="\n")


if __name__ == "__main__":
    main()



"""Helper functions."""
from typing import List

from pyimgtool.data_structures import Config, Context


def humanize_bytes(num, suffix="B", si_prefix=False, round_digits=2) -> str:
    """Return a human friendly byte representation.

    Modified from: https://stackoverflow.com/questions/1094841/1094933#1094933
    """
    div = 1000.0 if si_prefix else 1024.0
    unit_suffix = "i" if si_prefix else ""
    for unit in ["", "k", "M", "G", "T", "P", "E", "Z"]:
        if abs(num) < div:
            return f"{num:3.{round_digits}f} {unit}{unit_suffix}{suffix}"
        num /= div
    return f"{num:3.{round_digits}f} Y{unit_suffix}{suffix}"


def get_summary_report(cfg: Config, ctx: Context) -> List[str]:
    """Create formatted report based on details in `cfg` and `ctx`.

    Parameters
    ----------
    - `cfg` Config object
    - `ctx` Context object

    Returns
    -------
    - List of strings of formatted report "table"

    """
    size_reduction_bytes = ctx.orig_file_size - ctx.new_file_size
    report_title = " Processing Summary "
    report_end = " End "
    report_arrow = "->"
    report = []
    report.append(["File Name:", cfg.input_file, report_arrow, cfg.output_file])
    report.append(
        ["File Dimensions:", str(ctx.orig_size), report_arrow, str(ctx.new_size)]
    )
    report.append(
        [
            "File Size:",
            humanize_bytes(ctx.orig_file_size),
            report_arrow,
            humanize_bytes(ctx.new_file_size),
        ]
    )
    report.append(
        [
            "Size Reduction:",
            f"{humanize_bytes(size_reduction_bytes)} "
            f"({(size_reduction_bytes/ctx.orig_file_size) * 100:2.1f}%)",
        ]
    )
    report.append(
        ["Processing Time:", f"{(ctx.time_end - ctx.time_start)*1000:.1f} ms"]
    )
    for c in report:
        for n in range(4):
            try:
                c[n] = c[n]
            except IndexError:
                c.append("")
        c[2] = "" if c[3] == c[1] else c[2]
        c[3] = "  " if c[3] == c[1] else c[3]

    padding = 2
    col0w = max([len(str(c[0])) for c in report]) + padding
    col1w = max([len(str(c[1])) for c in report]) + padding
    col2w = max([len(str(c[2])) for c in report]) + padding
    col3w = max([len(str(c[3])) for c in report]) + padding
    out = []
    out.append(f"{report_title:{'-'}^{col0w + col1w + col2w + col3w + 1}}")
    for line in report:
        out.append(
            f"{line[0]:<{col0w}} {line[1]:{col1w}} {line[2]:{col2w}} {line[3]:{col3w}}"
        )
    out.append(f"{report_end:{'-'}^{col0w + col1w + col2w + col3w + 1}}")
    return out



"""Image operations."""

import logging
import sys
from io import BytesIO

import piexif
from PIL import Image

from pyimgtool import resize, watermark
from pyimgtool.data_structures import Config, Context
from pyimgtool.utils import humanize_bytes

LOG = logging.getLogger(__name__)


def calculate_new_size(cfg: Config, ctx: Context) -> None:
    """Update Config with correct width/height.

    Percent scale (`-p`) takes precedence over width (`-mw`) and
    height (`-mh`). Func does nothing if both height and width
    are supplied at the command line.

    Parameters
    ----------
    - `cfg` Config object
    - `ctx` Context object

    """
    if cfg.pct_scale:
        LOG.info("Scaling image by %.1f%%", cfg.pct_scale)
        cfg.width = int(round(ctx.orig_size.width * (cfg.pct_scale / 100.0)))
        cfg.height = int(round(ctx.orig_size.height * (cfg.pct_scale / 100.0)))
    elif cfg.width and not cfg.height:
        LOG.info("Calculating height based on width")
        cfg.height = int(
            round((cfg.width * ctx.orig_size.height) / ctx.orig_size.width)
        )
    elif cfg.height and not cfg.width:
        LOG.info("Calculating width based on height")
        cfg.width = int(
            round((cfg.height * ctx.orig_size.width) / ctx.orig_size.height)
        )
    elif not cfg.height and not cfg.width:
        LOG.info("No new width or height supplied; using current dims")
        cfg.width = ctx.orig_size.width
        cfg.height = ctx.orig_size.height


def process_image(cfg: Config) -> Context:
    """Process image according to options in `cfg`."""
    ctx = Context()
    inbuf = BytesIO()
    outbuf = BytesIO()
    if not cfg.input_file:
        raise ValueError("input_file required")
    with open(cfg.input_file, "rb") as f:
        inbuf.write(f.read())
    ctx.orig_file_size = inbuf.tell()
    im = Image.open(inbuf)
    try:
        exif = piexif.load(im.info["exif"], True)
        del exif["thumbnail"]
        ctx.orig_exif = exif
    except KeyError:
        pass
    ctx.orig_size.width, ctx.orig_size.height = im.size
    ctx.orig_dpi = im.info["dpi"]
    LOG.info("Input dims: %s", ctx.orig_size)
    LOG.info("Input size: %s", humanize_bytes(ctx.orig_file_size))

    calculate_new_size(cfg, ctx)

    if cfg.watermark_image is not None:
        im = watermark.with_image(im, cfg, ctx)
    if cfg.text is not None or cfg.text_copyright is not None:
        im = watermark.with_text(im, cfg, ctx)

    # Resize/resample
    if cfg.height != ctx.orig_size.height or cfg.width != ctx.orig_size.width:
        im = resize.resize_thumbnail(
            im,
            (cfg.width, cfg.height),
            #  bg_size=(cfg.width + 50, cfg.height + 50),
            resample=Image.ANTIALIAS,
        )

    try:
        ctx.new_dpi = im.info["dpi"]
    except KeyError:
        pass
    LOG.info("Image mode: %s", im.mode)

    # Save
    im.save(outbuf, "JPEG", quality=cfg.jpg_quality, dpi=ctx.orig_dpi, progressive=True)
    ctx.image_buffer = outbuf.getvalue()

    # convert back to image to get size
    if ctx.image_buffer:
        img_out = Image.open(BytesIO(ctx.image_buffer))
        ctx.new_size.width, ctx.new_size.height = img_out.size
        ctx.new_file_size = sys.getsizeof(ctx.image_buffer)
    LOG.info("Output size: %s", humanize_bytes(ctx.new_file_size))
    return ctx



"""Configuration and argument parsing."""

import argparse
import logging
import textwrap
from pathlib import Path

from pyimgtool.data_structures import Position
from pyimgtool.version import __version__

LOG = logging.getLogger(__name__)


def parse_args(args: list):
    """Parse command line arguments."""
    # flags
    desc = textwrap.dedent(
        """\
        A command-line utility which uses the Pillow module to
        manipulate images for web vewing.

        Images can be resampled, resized, and compressed at custom
        quality levels. Watermarking can also be added.
        """
    )
    parser = argparse.ArgumentParser(
        prog="pyimgtool",
        description=desc,
        add_help=False,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    # Positionals
    parser.add_argument("input", help="image file to process", metavar="INPUT")
    parser.add_argument("output", help="file to save processed image", metavar="OUTPUT")

    # Flags
    parser.add_argument(
        "-c",
        help="read from this config file (default: conf.ini)",
        dest="config_file",
        default="../conf.ini",
    )
    parser.add_argument(
        "-f", help="force overwrite of existing file", dest="force", action="store_true"
    )
    parser.add_argument(
        "-v",
        help="increase logging output to console",
        action="count",
        dest="verbosity",
        default=0,
    )
    parser.add_argument("-h", action="help", help="show this help message and exit")
    parser.add_argument("--help", action="help", help=argparse.SUPPRESS)
    parser.add_argument("-V", action="version", version=f"%(prog)s {__version__}")
    parser.add_argument(
        "-s",
        nargs=1,
        help="text suffix appended to INPUT path if no OUTPUT file given",
        metavar="TEXT",
        dest="suffix",
        default="_edited",
    )
    parser.add_argument(
        "-n",
        help="display results only; don't save file",
        dest="no_op",
        action="store_true",
    )
    parser.add_argument(
        "-qu",
        help="quiet debug log output to console (opposite of -v)",
        action="store_true",
        dest="quiet",
    )

    # Image group
    image_group = parser.add_argument_group("General image options")
    image_group.add_argument(
        "-p",
        help="scale output by percent of orig size",
        dest="pct_scale",
        metavar="SCALE",
        type=float,
    )
    image_group.add_argument(
        "-mw",
        help="maximum width of output",
        dest="width",
        metavar="WIDTH",
        type=int,
        default=0,
    )
    image_group.add_argument(
        "-mh",
        help="maximum height of output",
        dest="height",
        metavar="HEIGHT",
        type=int,
        default=0,
    )
    image_group.add_argument(
        "-ke", help="keep exif data if possible", dest="keep_exif", action="store_true"
    )

    # Watermark group
    watermark_group = parser.add_argument_group("Watermark options")
    watermark_group.add_argument(
        "-wi",
        help="image file to use as watermark",
        type=Path,
        dest="watermark_image",
        metavar="PATH",
    )
    watermark_group.add_argument(
        "-wr",
        help="angle of watermark rotation",
        dest="watermark_rotation",
        metavar="ANGLE",
        type=int,
        default=0,
    )
    watermark_group.add_argument(
        "-wo",
        help="watermark opacity",
        dest="watermark_opacity",
        type=float,
        metavar="OPACITY",
        default=0.3,
    )
    watermark_group.add_argument(
        "-wp",
        help="watermark position",
        dest="watermark_position",
        metavar="POS",
        default=Position.BOTTOM_RIGHT,
        type=Position.argparse,
        choices=list(Position),
    )
    watermark_group.add_argument(
        "-ws",
        help="watermark scale in percent of image size (default = 10)",
        dest="watermark_scale",
        metavar="SCALE",
        default=0.2,
        type=float,
    )
    text_group = parser.add_argument_group("Text options")
    text_group.add_argument(
        "-t", help="text to display on image", dest="text", metavar="TEXT", type=str
    )
    text_group.add_argument(
        "-tc",
        help="display copyright message after © and date taken",
        dest="text_copyright",
        metavar="MESSAGE",
        type=str,
    )
    text_group.add_argument(
        "-tr",
        help="angle of text rotation",
        dest="text_rotation",
        metavar="ANGLE",
        type=int,
        default=0,
    )
    text_group.add_argument(
        "-to",
        help="text opacity",
        dest="text_opacity",
        type=float,
        metavar="OPACITY",
        default=0.3,
    )
    text_group.add_argument(
        "-tp",
        help="text position",
        dest="text_position",
        metavar="POS",
        default=Position.BOTTOM_RIGHT,
        type=Position.argparse,
        choices=list(Position),
    )
    text_group.add_argument(
        "-ts",
        help="scale of text relative to image width",
        dest="text_scale",
        metavar="SCALE",
        default=0.20,
        type=float,
    )

    # Jpg group
    jpg_group = parser.add_argument_group("Jpeg options")
    jpg_group.add_argument(
        "-q",
        help="Quality setting for jpeg files (an integer between 1 and 100; default: 75)",
        type=int,
        dest="jpg_quality",
        default=75,
        metavar="QUALITY",
    )

    if parser._positionals.title is not None:
        parser._positionals.title = "Arguments"
    if parser._optionals.title is not None:
        parser._optionals.title = "Options"
    parsed = parser.parse_args(args)

    # do basic validation
    if parsed.quiet > 0:
        parsed.verbosity = 0

    if not 0 <= parsed.jpg_quality <= 100:
        parser.error(f"Quality (-q) must be within 0-100; found: {parsed.jpg_quality}")

    if parsed.pct_scale and (parsed.width or parsed.height):
        parser.error("Can use either -p or -mw/-mh, not both")

    if not 0 <= parsed.watermark_scale <= 1:
        parser.error("Value out of bounds: -ws must be between 0 and 1")

    if parsed.text is not None and parsed.text_copyright is not None:
        parser.error("Can use either -t or -tc, not both")

    return parsed



"""Define application dependencies and entry points."""
from setuptools import setup, find_packages
from os import path

cwd = path.abspath(path.dirname(__file__))


def find_version(file):
    """Read version number from source file."""
    version = {}
    try:
        with open(path.join(cwd, file), "r") as f:
            version_file = f.read()
    except Exception:
        raise RuntimeError(f"Unable to open '{file}' to get version.")
    exec(version_file, version)

    try:
        version_str = version["__version__"]
    except IndexError:
        raise RuntimeError(f"Unable to get version from '{file}'")
    return version_str


def get_long_description(readme_file):
    """Read long description from README file."""
    try:
        with open(path.join(cwd, readme_file), "r", encoding="utf-8") as fh:
            return fh.read()
    except Exception:
        return ""


setup(
    name="pyimgtool",
    version=find_version("pyimgtool/version.py"),
    author="Nick Murphy",
    author_email="comfortablynick@gmail.com",
    description="Tool to help prep images for web sharing",
    long_description=get_long_description("README.md"),
    long_description_content_type="text/markdown",
    url="https://github.com/comfortablynick/pyimgtool",
    packages=find_packages(),
    include_package_data=True,
    install_requires=["pillow", "piexif", "attrs", "sty"],
    entry_points={"console_scripts": ["pyimgtool = pyimgtool.__main__:main"]},
    python_requires=">=3.6",
    project_urls={
        "Bug Reports": "https://github.com/comfortablynick/pyimgtool/issues",
        "Source": "https://github.com/comfortablynick/pyimgtool",
    },
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
    ],
)



"""Resize and validation functions.

Based on original code from:
https://github.com/VingtCinq/python-resize-image
"""
import logging
import math
import sys
from functools import wraps
from typing import Tuple

from PIL import Image

from pyimgtool.exceptions import ImageSizeError

LOG = logging.getLogger(__name__)


def validate(validator):
    """Return a decorator that validates arguments with provided `validator` function.

    This will also store the validator function as `func.validate`.
    The decorator returned by this function, can bypass the validator
    if `validate=False` is passed as argument otherwise the fucntion is
    called directly.

    The validator must raise an exception, if the function can not
    be called.
    """

    def decorator(func):
        """Bound decorator to a particular validator function."""

        @wraps(func)
        def wrapper(image, size, validate=True):
            if validate:
                validator(image, size)
            return func(image, size)

        return wrapper

    return decorator


def _is_big_enough(image: Image, size: Tuple[int, int]):
    """Check that the image's size superior to `size`."""
    if (size[0] > image.size[0]) and (size[1] > image.size[1]):
        raise ImageSizeError(image.size, size)


def _width_is_big_enough(image: Image, width: int):
    """Check that the image width is superior to `width`."""
    if width > image.size[0]:
        raise ImageSizeError(image.size[0], width)


def _height_is_big_enough(image: Image, size: Tuple[int, int]):
    """Check that the image height is superior to `height`.

    Raise ImageSizeError if image height < `height`.

    Parameters
    ----------
    - `image` PIL image
    - `height` Integer of image height

    """
    if size[1] > image.size[1]:
        raise ImageSizeError(image.size[1], size[1])


@validate(_is_big_enough)
def resize_crop(image: Image, size: Tuple[int, int]) -> Image:
    """Crop the image with a centered rectangle of the specified size.

    Parameters
    ----------
    - `image` Pillow image instance
    - `size` 2-Tuple of image dimensions (width, height)

    Returns
    -------
    - PIL Image

    """

    img_format = image.format
    image = image.copy()
    old_size = image.size
    left = (old_size[0] - size[0]) / 2
    top = (old_size[1] - size[1]) / 2
    right = old_size[0] - left
    bottom = old_size[1] - top
    rect = [int(math.ceil(x)) for x in (left, top, right, bottom)]
    left, top, right, bottom = rect
    crop = image.crop((left, top, right, bottom))
    crop.format = img_format
    return crop


@validate(_is_big_enough)
def resize_cover(image: Image, size: Tuple[int, int], resample=Image.LANCZOS) -> Image:
    """Resize image to fill the specified area; crop as needed.

    Same behavior as `background-size: cover`.

    Parameters
    ----------
    - `image` Pillow image instance
    - `size` 2-Tuple of image dimensions (width, height)
    - `resample` Resample method

    Returns
    -------
    - PIL Image

    """
    img_format = image.format
    img = image.copy()
    img_size = img.size
    ratio = max(size[0] / img_size[0], size[1] / img_size[1])
    new_size = [
        int(math.ceil(img_size[0] * ratio)),
        int(math.ceil(img_size[1] * ratio)),
    ]
    img = img.resize((new_size[0], new_size[1]), resample)
    img = resize_crop(img, size)
    img.format = img_format
    return img


def resize_contain(
    image: Image,
    size: Tuple[int, int],
    resample=Image.LANCZOS,
    bg_color: Tuple[int, int, int, int] = (255, 255, 255, 0),
    bg_size: Tuple[int, int] = None,
) -> Image:
    """Resize image to fill specified area.

    Image is not cropped and aspect ratio is kept intact.
    Same behavior as `background-size: contain`.

    Parameters
    ----------
    - `image` Pillow image instance
    - `size` 2-Tuple of image dimensions (width, height)
    - `resample` Resample method
    - `bg_color` RGBA Tuple for background (if image smaller than `size`)
    - `bg_size` Background size (if different from `size`)

    Returns
    -------
    - PIL Image

    """
    img_format = image.format
    img = image.copy()
    img.thumbnail((size[0], size[1]), resample)
    if not bg_size:
        bg_size = size
    background = Image.new("RGBA", bg_size, bg_color)
    img_position = (
        int(math.ceil((bg_size[0] - img.size[0]) / 2)),
        int(math.ceil((bg_size[1] - img.size[1]) / 2)),
    )
    background.paste(img, img_position)
    background.format = img_format
    return background.convert("RGB")


@validate(_width_is_big_enough)
def resize_width(image: Image, size: Tuple[int, int], resample=Image.LANCZOS) -> Image:
    """Resize image to according to specified width.

    Aspect ratio is kept intact.

    Parameters
    ----------
    - `image` Pillow image instance
    - `size` 2-Tuple of dimension integers
    - `resample` Resample method

    Returns
    -------
    - PIL Image

    """
    img_format = image.format
    img = image.copy()
    width = size[0]
    # If the origial image has already the good width, return it
    if img.width == width:
        return image
    new_height = int(math.ceil((width / img.width) * img.height))
    img.thumbnail((width, new_height), resample)
    img.format = img_format
    return img


@validate(_height_is_big_enough)
def resize_height(image: Image, size: Tuple[int, int], resample=Image.LANCZOS) -> Image:
    """Resize image to according to specified height.

    Aspect ratio is kept intact.

    Parameters
    ----------
    - `image` Pillow image instance
    - `size` 2-Tuple of dimension integers
    - `resample` Resample method

    Returns
    -------
    - PIL Image

    """
    img_format = image.format
    img = image.copy()
    # If the origial image has already the good height, return it
    height = size[1]
    if img.height == height:
        return image
    new_width = int(math.ceil((height / img.height) * image.width))
    img.thumbnail((new_width, height), resample)
    img.format = img_format
    return img


def resize_thumbnail(
    image: Image, size: Tuple[int, int], resample=Image.LANCZOS
) -> Image:
    """Resize image to according to specified size.

    Aspect ratio is kept intact while trying best to match `size`.

    Parameters
    ----------
    - `image` Pillow image instance
    - `size` 2-Tuple of dimension integers
    - `resample` Resample method

    Returns
    -------
    - PIL Image

    """
    img_format = image.format
    img = image.copy()
    img.thumbnail((size[0], size[1]), resample)
    img.format = img_format
    return img


def resize(method, *args, **kwargs):
    """Direct arguments to one of the resize functions.

    Parameters
    ----------
    - `method` One among 'crop', 'cover', 'contain', 'width', 'height' or 'thumbnail'
    - `image` Pillow image instance
    - `size` 2-tuple of integers [width, height]

    """
    valid_methods = ["crop", "cover", "contain", "width", "height", "thumbnail"]
    if method not in valid_methods:
        raise ValueError(
            f"method argument should be one of: {', '.join([ repr(m) for m in valid_methods])}"
        )
    method = f"resize_{method}"
    LOG.info("Resizing with %s()", method)
    return getattr(sys.modules[__name__], method)(*args, **kwargs)
